/* JISSA- Generic Disassembler for Java-Bytecode
 
 config.c++
 Routines for reading the config file.

 Copyright (c) 2000 Matthias Kramm <kramm@iname.com> 

 This file is distributed under the GPL, see file COPYING for details */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <map.h>

char basepath[128]; // path to the executable/ local directory

extern FILE*output; 
extern int verbose;

int spec=0;
char space[1024];	// buffer for filename creation

/*
 * Impementation of string comparison for the STL map class
 */

struct str
{
	char*s;
	public:
	str() {s=0;};
	str(char*_s) {s=_s;}
	bool operator ==(str t) {return !strcmp(s,t.s);}
};

class less<str>
{
	public:
	bool operator()(str x,str y) const 
	{ return strcmp(x.s,y.s)<0;}
};

class equal_to<char*>
{
	public:
	bool operator()(str x,str y) const 
	{ return !strcmp(x.s,y.s);}
};

class not_equal_to<char*>
{
	public:
	bool operator()(str x,str y) const 
	{ return strcmp(x.s,y.s);}
};

map<str,str> table;

/**
 * void assign(char*key,char*value)
 *
 * key		the key the value is assigned to
 * value	the value to assign
 *
 *
 */
void assign(char*key,char*value)
{
	int l1=strlen(key);
	int l2=strlen(value);
	char*a1=new char[l1+1];
	char*a2=new char[l2+1];
	strcpy(a1,key);
	strcpy(a2,value);
	str s1(a1);
	str s2(a2);

	table[s1]=s2;
}

/**
 * char* getval(char*key)
 *
 * get's the value from key
 *
 * key		the value to get
 *
 */
char* getval(char*key)
{
	str s(key);
	char*a=table[s].s;

	if(a) return a;
	else return "";
}

char includepath[2048];
int includepathpos=0;
/**
 * void includefiles_append(char*dir)
 *
 * appends a given string to the list of include files
 *
 * name			a ;-seperated string of directories
 */
void includefiles_append(char*dir)
{
	if(';'!=dir[0] && ';'!=includepath[includepathpos-1] && includepathpos>0)
		includepath[includepathpos++]=';';
	strcpy(&includepath[includepathpos],dir);
	includepathpos+=strlen(dir);
	includepath[includepathpos]=';';
}

/** 
 * void tryopen(char*name)
 *
 * try to open a file under the given name, taking account the possiblity 
 * the extension or the path has been left out.
 */

FILE* tryopen(char*name,int include)
{
	FILE*fi;
	char*ip=includepath;
	while(ip-includepath<includepathpos)
	{
		int len=strchr(ip,';')-ip;
		strncpy(space,ip,len);
		ip+=len+1;
		if(space[len-1]!='/')
		space[len++]='/';
		strcpy(&space[len],name);
		fi=fopen(space,"r");
		if(verbose) fprintf(stderr,"Trying %s\n",space);
		if(fi) return fi;
		if(include) strcat(space,".inc");
		else		strcat(space,".fmt");
		fi=fopen(space,"r");
		if(verbose) fprintf(stderr,"Trying %s\n",space);
		if(fi) return fi;
	}
	return 0;
}
void readaline(FILE*fi,char*t)
{
    *t=0;
    while(1)
    {
	if(!fread(t,1,1,fi)) {*t=0;return;}
	if(*t==10) {*t=0;break;}
	if(*t==13) {*t=0;break;}
	t++;
    }
    return;
}
/**
 * void readconfig2(char*name,int include=0)
 *
 * reads a config file
 *
 * name			the name of the config file
 * include		1 if the file is the include of another config file
 */
void readconfig2(char*name,int include=0)
{
	char*c=new char[65536];
	char*d=new char[65536];
	char*e=new char[65536];

	int line=0;

	FILE*fi;

	fi=fopen(name,"r");

	if(!fi) //TRY1
	{
	    if(verbose) fprintf(stderr,"%s not found\n",name);
	    fi=tryopen(name,include);
	    
	    if(!fi)
	    {
		if(include) fprintf(stderr,"Could not find include file %s!\n",name);
		else		
		  fprintf(stderr,"Could not find config file %s! (jissa --help-config)\n",name);
		exit(1);
	    }
	    if(verbose) fprintf(stderr,"Found %s\n",space);
	}
	
	while(!feof(fi))
	{
		//if(!fscanf(fi,"%[^\n]\n",c)) break;
		readaline(fi,c);
		line++;

		if(c[0]=='*') {spec^=1;continue;}

		d[0]=0;
		e[0]=0;
		if(c[0]==0) continue;
		int l=sscanf(c," %s %[^\n]\n",d,e);
		if(d[0]=='#') continue;

		if(!strcasecmp(d,"include"))
		{
			if(l!=2)
			{
				fprintf(stderr,"include with no params found in file %s line %d\n",
				name,line);
				exit(1);
			}
			readconfig2(e,1);
		}

		if(!spec)
		{
			fprintf(output,"%s\n",c);
		}
		else
		{
			if(l!=2) 
			{
				fprintf(stderr,"[%d] Error in config file line %d\n",l,line);
				exit(1);
			}
			assign(d,e);
		}
	}

	fclose(fi);

	delete c;
	delete d;
	delete e;
}

/**
 * void readconfig(char*name)
 *
 * reads a config file
 *
 * name			the name of the config file
 */
void readconfig(char*name)
{
	spec=0;
	readconfig2(name);
}

