/* JISSA- Generic Disassembler for Java-Bytecode
 
 disass.c++
 implements main()

 Copyright (c) 2000 Matthias Kramm <kramm@iname.com> 

 This file is distributed under the GPL, see file COPYING for details */

#include "stdlib.h"
#include "stdio.h"
#define HEADERS
#include "split.c++"
#include "config.c++"
#include "ops.c++"
#undef HEADERS

int verbose;

/**
 *
 *		global Variables
 *
*/
#define POOLSIZE 1048576

char*	fname;			// filename of the classfile
FILE*	fi=0;			// filedescriptor of the classfile
FILE*	output=0;		// filedescriptor of the intermediate perlfile
int		pass=0;			// pass, either 0 or 1. (every class gets parsed twice)
int		lpass=1;		// "local pass", 0 or 1, for methods which get parsed twice
int		fpos=0;			// the file position in fi
char*	mem;			// memory for temporary input
char*	datapool=0;		// constant pool size (in bytes)
int		poolpos;		// constant pool size (in entries)
int*	constantdatas=0;// data structure for the constant pool
int*	constanttype=0; // data structure for the types of the constant pool entries

/**
 * struct localvar
 *
 * data structure for the local variables table found in some classfiles
 *
 */
struct localvar
{
	u16 start_pc;
	u16 length;
	u16 name_index;
	u16 type_index;
	u16 slot;
} local_vars[65536];
int locals_count;

/**
 *
 * data structures for the linenumbertable found in some classfiles
 *
 */
u16 start_pc[65536];
u16 line_number[65536];
u16 lines_count;

/**
 *
 * data structure for try/catch blocks
 *
 */
struct handler_type {
		u16 start_pc;
		u16 end_pc;
		u16 handler_pc;
		u16 catch_type_index;
		} *handlers; 
u16 handlers_count;
/**
 *
 * bytecodeblock preprocessing data
 *
 */
u32*oppos;
u16*labels;
s32*cases;
u8*casetags;
u16 label_pos;
int code_size;

#define ID_CONSTANT_Utf8 1
#define ID_CONSTANT_Integer 3
#define ID_CONSTANT_Float 4
#define ID_CONSTANT_Long 5
#define ID_CONSTANT_Double 6
#define ID_CONSTANT_Class 7
#define ID_CONSTANT_String 8
#define ID_CONSTANT_Fieldref 9
#define ID_CONSTANT_Methodref 10
#define ID_CONSTANT_InterfaceMethodref 11
#define ID_CONSTANT_NameAndType 12

/**
 *
 *		Prototypes
 *
*/

int onedescriptor(char*str);
void splitdescriptor(char*str);

/**
 *
 *		System/in- output functions
 *
 */

/*
 * void error(char*e)
 * Prints an error-message and exists
 *
 * e Description of the Error
 *
 */
void error(char*e)
{
	fprintf(stderr,"%s\n",e);
	exit(0);
}

/*
 * void input(void* p,int l)
 * Read a number bytes out of the input-classfile
 *
 * p  the memory location where the data should go
 * l  the the number of bytes to read
 *
 */
void input(void*p,int l)
{
	if(l)
	{
		if(!fread(p,l,1,fi)) 
			error("Unexpected end of file!\n");
		fpos+=l;
	}
}

/*
 * void input1(void* p)
 * Read one byte out of the input-classfile
 *
 * p  the memory location where the data should go
 *
 */
void input1(void*p)
{
	input(p,1);
}

/*
 * void input2(void* p)
 * Read two bytes (one Word) out of the input-classfile. Performs neccessary
 * Big-Endian -> Little Endian conversions
 *
 * p  the memory location where the data should go
 *
 */
void input2(void*p)
{
	input(p,2);
	reverse2(p);
}

/*
 * void input4(void* p)
 * Read four bytes (one Doubleword/float) out of the input-classfile. 
 * Performs neccessary Big-Endian -> Little Endian conversions
 *
 * p  the memory location where the data should go
 *
 */
void input4(void*p)
{
	input(p,4);
	reverse4(p);
}

/*
 * void input8(void* p)
 * Read eight bytes (one Doubleword/float) out of the input-classfile. 
 * Performs neccessary Big-Endian -> Little Endian conversions
 *
 * p  the memory location where the data should go
 *
 */
void input8(void*p)
{
	input(p,4);
	reverse8(p);
}

/*
 * void get1(void* p)
 *
 * Returns the byte at the given memory location
 *
 * *p the memory location
 *
 */
u8 inline get1(void*p)
{
    return *(u8*)p; 
} 

/*
 * void get2(void* p)
 *
 * Returns the word at the given memory location
 *
 * *p the memory location
 *
 */
u16 get2(void*p)
{
	u16 a=(int)((u8*)p)[1]+(int)((u8*)p)[0]*256;
	return a;
}

/*
 * void get4(void* p)
 *
 * Returns the doubleword at the given memory location
 *
 * *p the memory location
 *
 */
u32 get4(void*p)
{
	u32 a=(u32)((u8*)p)[3]+(u32)((u8*)p)[2]*256
	       +(u32)((u8*)p)[1]*65536+(u32)((u8*)p)[0]*256*65536;
	return a;
}

/*
 * void inputmem1(void*p)
 *
 * Like input1, but reads from memory (used for nested data structures)
 *
 * *p the memory location
 *
 */
void inputmem1(void*p)
{
	*(u8*)p=*(u8*)mem;
	mem++;
}

/*
 * void inputmem2(void*p)
 *
 * Like input2, but reads from memory (used for nested data structures)
 *
 * *p the memory location
 *
 */
void inputmem2(void*p)
{
	*(u16*)p=(int)((u8*)mem)[1]+(int)((u8*)mem)[0]*256;
	mem+=2;
}

/*
 * void inputmem4(void*p)
 *
 * Like input4, but reads from memory (used for nested data structures)
 *
 * *p the memory location
 *
 */
void inputmem4(void*p)
{
	*(u32*)p= (u32)((u8*)mem)[3]+(u32)((u8*)mem)[2]*256
	              +(u32)((u8*)mem)[1]*65536+(u32)((u8*)mem)[0]*256*65536;
	mem+=4;
}

/*
 * void inputmem8(void*p)
 *
 * Like input8, but reads from memory (used for nested data structures)
 *
 * *p	the memory location
 *
 */
void inputmem8(void*p)
{
	int t;
	for(t=0;t<8;t++) 
		((u8*)p)[t]=((u8*)mem)[t];
	reverse8(p);
	mem+=8;
}


/*
 * void printout(char*tag)
 *
 * takes a string of the form >tag< and converts it to
 * >printf tag<, which is written to the temporary outputfile 
 * (/tmp/output.pl)
 *
 * tag	the string 
 *
 */
void printout(char*tag)
{
	if(pass && lpass)
	{
		fprintf(output,"\n#%s\n",tag);
		fprintf(output,"printf %s;\n",getval(tag));
		fflush(output);
	}
	else
	{
		fprintf(output,"\n");
	}
}

/*
 * void printout(char*tag)
 *
 * takes a string of the form >tag< and a variable name and
 * converts it to >$var=tag<, which is written to the temporary outputfile
 * Used for constant-pool-conversions
 *
 * tag	the string to be assigned
 * var	the variable to assign to
 *
 */
void storevar(char*var,char*tag)
{
	if(pass)
	{
		char*s=getval(tag);
		if(s && *s)
		fprintf(output,"%s=%s;\n",var,getval(tag));
	}
	else
	{
	}
}

/*
 * void inputp(void*p,int len)
 *
 * like input, but stores the data which was read into the local constant
 * pool structure.
 *
 * p	the memory to read to
 * len	the number of bytes to read 
 *
 */
void inputp(void*p,int len)
{
	input(p,len);
	memcpy(&datapool[poolpos],p,len);
	poolpos+=len;
}

/*
 * void readutf8p(int len)
 *
 * reads an utf8-string per inputp()
 *
 * len	the length of the string
 * memlen	0: len is the number of real characters
 *			1: len is the number of bytes the string occupeis
 *
*/
void readutf8p(int len,int memlen=1)
{
	if(memlen)
	{
		int t;
		for(t=0;t<len;t++)
		{
			unsigned char b1;
			inputp(&b1,1);
		}
	}
	/*
		 according to o'reillys "Java Virtual Machine", first edition,
		 this is the way utf8 is encoded in java.

		 just in case len is the number of read characters and not the memory
		 length, we need to do a little more work
	*/
	else
	{
		int t;
		for(t=0;t<len;t++)
		{
			unsigned char b1;
			inputp(&b1,1);
			if(b1<128) continue;
			unsigned char b2;
			inputp(&b2,1);
			if(b1<192+32) continue;
			unsigned char b3;
			inputp(&b3,1);
		}
	}
}

/**
 *	void readconstantpool(int num)
 *
 *	read the constant pool out of the classfile
 *
 *	num		the number of constant-pool-entries (from classheader)
 *
 */
void readconstantpool(int num)
{
	if(!datapool)
		datapool=new char[POOLSIZE];

	poolpos=0;

    if(!constantdatas)
	constantdatas  =new int[num];
    if(!constanttype)
	constanttype   =new int[num];

	int t=1;
	char*name;
	int s;
	int sadd;
	for(s=1;s<num;s++) //num-1 entries!
	{
		sadd=0;
		if(s!=1)
		printout("constants_separator");
		u8 tag;
		input(&tag,1);

		constanttype[t]=tag;
		constantdatas[t]=poolpos;

		fprintf(output,"$i=%d;",s);

		switch(tag)
		{
		 case ID_CONSTANT_Integer:
			{
			
			u32 value;
			inputp(&value,4);
			reverse4(&value);
			fprintf(output,"$constantpool1[%d]=%d;\n",t,value);

			fprintf(output,"$val=%d;",value);
			storevar("$value","CONSTANT_Integer");
			break;
			}
			
		 case ID_CONSTANT_Float:
		 {
			u32 value;
			inputp(&value,4);
			reverse4(&value);

			char* num="",*denom="";
			int sign,mantissa,exp;
			//splitfloat(value,&num,&denom,&sign,&exp,&mantissa);
			fprintf(output,"$constantpool1[%d]=%f;\n",t,*(float*)&value);

			fprintf(output,"$val=%f;$numerator=\"%s\";$denominator=\"%s\";"
			               "$sign=%d;$mantissa=\"%d\";$exp=%d;",*(float*)&value,
								  num,denom,sign,mantissa,exp);
			storevar("$value","CONSTANT_Float");
			break;

		 }
		 case ID_CONSTANT_Long:
		 {
			u64 value;
			inputp(&value,8);
			reverse8(&value);

			char*val;
			splitlong(value,&val);
			fprintf(output,"$constantpool1[%d]=\"%s\";\n",t,val);

			fprintf(output,"$val=%s;",val);
			storevar("$value","CONSTANT_Long");
			sadd=1;
			break;

		 }
		 case ID_CONSTANT_Double:
		 {
			u64 value;
			inputp(&value,8);
			reverse8(&value);

			char* num,*denom;
			int sign,exp;
			u64 mantissa;

			splitdouble(value,&num,&denom,&sign,&exp,&mantissa);
			fprintf(output,"$constantpool1[%d]=%lf;\n",t,*(double*)&value);

			fprintf(output,"$val=%f;$numerator=\"%s\";$denominator=\"%s\";",
			*(double*)&value,num,denom);

            char* mantissa2;
			splitlong(mantissa,&mantissa2);
			fprintf(output,"$sign=%d;$mantissa=\"%s\";$exp=%d;",
								  sign,mantissa2,exp);
			storevar("$value","CONSTANT_Float");
			sadd=1;
			break;
		 }
		 case ID_CONSTANT_Utf8:
		 {
			u16 len;
			input(&len,2);
			reverse2(&len);

			readutf8p(len);
			datapool[poolpos++]=0;

			char*a=&datapool[constantdatas[t]];
			fprintf(output,"$constantpool1[%d]=\"%s\";\n",t,formatstr(a));

			fprintf(output,"$str=\"%s\";",formatstr(a));
			storevar("$value","CONSTANT_Utf8");
			break;

		 }
		 case ID_CONSTANT_Class: name="CONSTANT_Class"; goto STRINGINDEX;
		 case ID_CONSTANT_String: name="CONSTANT_String"; goto STRINGINDEX;
		 {

            STRINGINDEX:
			u16 index;
			inputp(&index,2);
			reverse2(&index);
			fprintf(output,"$index=%d;",index);	
			fprintf(output,"$constantpool1[%d]=%d;\n",t,(int)index);

			fprintf(output,"$str=$constantpool1[%d];",
			       index);
			storevar("$value",name);
			break;

		 }
		 case ID_CONSTANT_Fieldref: name="CONSTANT_Fieldref"; goto CLASSNT;
		 case ID_CONSTANT_Methodref: name="CONSTANT_Methodref"; goto CLASSNT;
		 case ID_CONSTANT_InterfaceMethodref: name="CONSTANT_InterfaceMethodref"; goto CLASSNT;
		 {

            CLASSNT:
			u16 class_index;
			inputp(&class_index,2);
			reverse2(&class_index);

			u16 nt_index;
			inputp(&nt_index,2);
			reverse2(&nt_index);

			fprintf(output,"$constantpool1[%d]=%d;\n",t,class_index);
			fprintf(output,"$constantpool2[%d]=%d;\n",t,   nt_index);

			fprintf(output,"$class_index=%d;$nt_index=%d;"
			        "$class_name=$constantpool1[$constantpool1[$class_index]];"
					"$name=$constantpool1[$constantpool1[$nt_index]];"
					"$type=$constantpool1[$constantpool2[$nt_index]];",
					class_index,nt_index);
			storevar("$value",name);
			break;
		 }
		 case ID_CONSTANT_NameAndType:
		 {
			u16 name_index;
			inputp(&name_index,2);
			reverse2(&name_index);

			u16 type_index;
			inputp(&type_index,2);
			reverse2(&type_index);

			fprintf(output,"$constantpool1[%d]=%d;\n",t,name_index);
			fprintf(output,"$constantpool2[%d]=%d;\n",t,type_index);

			fprintf(output,"$name_index=%d;$type_index=%d;"
			        "$name=$constantpool1[$name_index];"
					"$type=$constantpool1[$type_index];",
					name_index,type_index);
			storevar("$value","CONSTANT_NameAndType");
			break;
		 }

			default:
			fprintf(stderr,"[%02x] at %d (fpos:%d): ",(int)tag,t,fpos);
			error("Contant type not recognized!");
			break;
		}
		t++;
		t+=sadd;
		fprintf(output,"$xpool[$i]=$value;");

		printout("constant");
		s+=sadd;
	}

	fprintf(output,"@constantpool=@pool=@constantpool1;\n");
	fprintf(output,"@pool2=@constantpool2;\n");

	printout("constants_tail");
}

/**
 *	void readfields(int num)
 *
 *	Reads all the fields the class
 *
 *	num		the number of fields 
 *
 */
void readfields(int num)
{
	int t;
	for(t=0;t<num;t++)
	{
		if(t) printout("fields_separator");
		struct
		{
			u16 access_flags;
			u16 name_index;
			u16 type_index;
			u16 attributes_count;
		} head;
		
		input2(&head.access_flags);
		input2(&head.name_index);
		input2(&head.type_index);
		input2(&head.attributes_count);

     	fprintf(output,"$field_flags=$flags=%d;$ACC_PUBLIC=%d;"
			          "$ACC_PRIVATE=%d;"
			          "$ACC_PROTECTED=%d;"
					  "$ACC_STATIC=%d;"
					  "$ACC_FINAL=%d;"
					  "$ACC_VOLATILE=%d;"
					  "$ACC_TRANSIENT=%d;",
					       head.access_flags,
					       head.access_flags&1,
						   (head.access_flags>>1)&1,
						   (head.access_flags>>2)&1,
						   (head.access_flags>>3)&1,
						   (head.access_flags>>4)&1,
						   (head.access_flags>>6)&1,
						   (head.access_flags>>7)&1);

		fprintf(output,"$name_index=%d;$type_index=%d;$attributes_count=%d;",
		head.name_index,head.type_index,head.attributes_count);
		fprintf(output,"$constant_index=0;");

		int s;
		for(s=0;s<head.attributes_count;s++)
		{
			struct
			{
				u16 attribute_name_index;
				u32 bytes_count;
			} head2;

			input2(&head2.attribute_name_index);
			input4(&head2.bytes_count);

			char*data=new char[head2.bytes_count];
			input(data,head2.bytes_count);

			if(!strcmp("ConstantValue",&datapool[constantdatas[head2.attribute_name_index]]))
			{
				u16 constant_value_index=get2(data);
				fprintf(output,"$constant_index=%d;",constant_value_index);
				fprintf(output,"$constant=$constantpool1[$constant_index];");
			}
			else
				fprintf(stderr,"Unrecognized Fieldattribute: %s (%d bytes)\n",&datapool[constantdatas[head2.attribute_name_index]],head2.bytes_count);

			delete data;
		}
		char*type=&datapool[constantdatas[head.type_index]];
		char*name=&datapool[constantdatas[head.name_index]];
		onedescriptor(type);
		fprintf(output,"$type=\"%s\";",type);
		fprintf(output,"$name=\"%s\";",formatstr(name));
		printout("field");
	}
	printout("fields_tail");
}

/**
 *	void readattributes(int num)
 *
 *	Reads all the attributes the class
 *
 *	num		the number of attributes
 *
 */
void readattributes(int num)
{
	int t;
	for(t=0;t<num;t++)
	{
		if(t) printout("attributes_separator");

		struct 
		{
			u16 attribute_name_index;
			u32 bytes_count;
		}
		head;

		input2(&head.attribute_name_index);
		input4(&head.bytes_count);

		fprintf(output,"$i=%d;$name=$constantpool1[$i];",head.attribute_name_index);
			                    // ID_CONSTANT_Utf8
		fprintf(output,"$num=%d;",head.bytes_count);

		char*b=new char[head.bytes_count];
		input(b,head.bytes_count);

		int s;
		for(s=0;s<head.bytes_count;s++)
		 fprintf(output,"$b[%d]=%d;",s,b[s]);

		printout("attribute");

		//Sourcefile

		char*name=&datapool[constantdatas[head.attribute_name_index]];
		if(!strcmp("SourceFile",name))
		{
			u16 source_file_index;
			source_file_index=get2(&b[0]);
			//reverse2(&source_file_index);
			fprintf(output,"$i=%d;$source_name=$name=\"%s\";",
			source_file_index,
			&datapool[constantdatas[source_file_index]]);
			printout("source_file");
		}
		else
		if(!strcmp("InnerClasses",name))
		{
			fprintf(stderr,"Unrecognized Attribute: %s (%d bytes)\n",name,head.bytes_count);
		}
		else
		if(!strcmp("Deprecated",name))
		{
			fprintf(output,"$class_deprecated=1;");
		}
		else
		fprintf(stderr,"Unrecognized Attribute: %s (%d bytes)\n",name,head.bytes_count);

		delete b;
	}
	printout("attributes_tail");
}

/**
 *  void readinterfaces(int num)
 * 
 *  Reads the interfaces the class implements
 *  
 *  num		the number of interfaces the class implements
 *
 */
void readinterfaces(int num)
{
	int t;
	for(t=0;t<num;t++)
	{
		if(t) printout("interfaces_separator");

		u16 index;
		input(&index,2);
		reverse2(&index);

	    fprintf(output,"$i=%d;$index=%d;\n",t,index);
		fprintf(output,"$name=$constantpool[$constantpool[$index]];\n");
		printout("interface");
	}
	printout("interfaces_tail");
}

/**
 * void newcode(int num)
 *
 * called whenever a new code block begins
 *
 * num		the number of bytes the bytecode occupies
 *
 */
void newcode(int num)
{
	oppos	   =new u32[num];
	labels     =new u16[num]; memset(labels,0,num*2);
	cases      =new s32[num]; memset(cases,0,num*4);
	casetags   =new u8[num]; memset(casetags,0,num);
	label_pos=1;
	code_size=num;
}

/**
 * void freecode()
 *
 * frees the memory allocated by newcode()
 */
void freecode()
{
	delete oppos;
	delete labels;
	delete cases;
	delete casetags;
}

/**
 * int oplen(u8*mem,int pos)
 * 
 * preprocesses the opcode in mem[pos]. calculates the next bytecode-position
 * relative to pos, and stores label and switch-information
 *
 * mem		the general location of the bytecode-instructions of the 
 *			current method
 * pos		the position of the position of the bytecode to preprocess
 */
int oplen(u8*mem,int pos)
{
	int opos=pos;
	int wide=0;
	if(mem[pos]==0xc4)
	{
		pos++;wide=1;
	}

	u8 op=mem[pos];
	char*form=format[mem[pos]];
	pos++;

	int j=1;
	while(*form)
	{
			int i=0;
			char c=*form;
			if(wide&&j)  
				c='w';
			j=0;
			switch(c)
			{
				case 'b': pos++;  break;
				case 'c': pos++;  break;
				case 'w': pos+=2; break;
				case 's': pos+=2; break;

				case 'j': 
				{
					s16 l=get2(&mem[pos]);
					u32 jmpadr=opos+l;
					if(!labels[jmpadr] && (jmpadr)<code_size && (jmpadr)>=0)
					 labels[jmpadr]=label_pos++;
					pos+=2; 
					break;
				}
				case 'J': 
				{
					u32 l=get4(&mem[pos]);
					if(!labels[l] && l<code_size && l>=0)
					 labels[l]=label_pos++;
					pos+=4; 
					break;
				}

			    case '!': 
				{
					if(op==0xaa) //tableswitch
					{
						while(pos&3) pos++; // padding
					    u32 jdefault=get4(&mem[pos]);
						if(!labels[jdefault+opos])
						 labels[jdefault+opos]=label_pos++;
						if(!casetags[jdefault+opos])
						 casetags[jdefault+opos]=2;
						pos+=4;
						s32 low=get4(&mem[pos]);pos+=4;
					    s32 high=get4(&mem[pos]);pos+=4;
						s32 t;
						for(t=low;t<=high;t++)
						{
							u32 l=get4(&mem[pos]);
							if(!labels[l+opos])
							 labels[l+opos]=label_pos++;
							if(!casetags[l+opos])
							{
							 cases[l+opos]=t;
							 casetags[l+opos]=1;
							}
							pos+=4;
						}
					}
					if(op==0xab) //lookupswitch
					{
						while(pos&3) pos++;
					    u32 jdefault=get4(&mem[pos]);
						
						if(!labels[jdefault+opos])
						 labels[jdefault+opos]=label_pos++;
						if(!casetags[jdefault+opos])
						 casetags[jdefault+opos]=2;
						pos+=4;
						u32 n=get4(&mem[pos]);pos+=4;
						int t;
						for(t=0;t<n;t++)
						{
							u32 key=get4(&mem[pos]);pos+=4;
							s32 offset=get4(&mem[pos]);pos+=4;
							if(!labels[offset+opos])
							 labels[offset+opos]=label_pos++;
							if(!casetags[offset+opos])
							{
							 cases[offset+opos]=key;
							 casetags[offset+opos]=1;
							}
						}
					}
				    break;
				}
				    
			}
			form++;
	}
	return pos;
}

/**
 * void readcode(int lpass,u8*mem,int num)
 *
 * Reads a bytecode-block, performing the desired output in case lpass=pass=1
 *
 * lpass	the number of the local pass
 * mem		the memoryposition of the bytecodes
 * num		the size of the bytecode-block
 */
void readcode(int lpass,u8*mem,int num)
{
	newcode(num);
	if(lpass) printout("ops_head");

	int pos=0;
	int nr=0;
	while(pos<num)
	{
		oppos[nr]=pos;
		pos=oplen(mem,pos);
		nr++;
	}

	int numorders=nr;
	pos=0;
	nr=0;
	int linepos=0;
	while(pos<num)
	{
		if(lpass) if(pos!=0) printout("ops_separator");

		if(lpass) fprintf(output,"$pos=%d;$nr=%d;",pos,nr);

		while(start_pc[linepos+1]<=pos)
		{
			if(linepos==lines_count)
			{
				if(lines_count>0)
				{
					fprintf(stderr,"%d Broken Line-Number table(2)\n",start_pc[linepos]);
					lines_count=0;
				}
				break;
			}
			else
			linepos++;
		}

		if(lines_count)
		{
			if(lpass) fprintf(output,"$linenumber=\"%-4d\";",line_number[linepos]);
		}
		else
		{
			if(lpass) fprintf(output,"$linenumber=\"\";");
		}

		{
		    int t;
		    for(t=0;t<locals_count;t++)
		    {
			if(local_vars[t].start_pc==pos)
			{
				localvar loc=local_vars[t];
				if(lpass) fprintf(output,"$start_pc=%d;",loc.start_pc);
			if(lpass) fprintf(output,"$length=%d;",loc.length);
				if(lpass) fprintf(output,"$name_index=%d;",loc.name_index);
				if(lpass) fprintf(output,"$type_index=%d;",loc.type_index);
				if(lpass) fprintf(output,"$slot=%d;\n",loc.slot);
				if(lpass) fprintf(output,"$localname[%d]=$pool[%d];\n",loc.slot,loc.name_index);
				if(lpass) fprintf(output,"$localtype[%d]=$pool[%d];\n",loc.slot,loc.type_index);

				if(lpass) printout("local");
			}
		    }

		    for(t=0;t<handlers_count;t++)
		    {
			u16 start_pc;
			u16 end_pc;
			u16 handler_pc;
			u16 catch_type_index;
		    
			if(handlers[t].start_pc == pos ||
			    handlers[t].end_pc == pos ||
			    handlers[t].handler_pc == pos)	
			{
			    if(lpass) fprintf(output,"$start_pc=%d;",handlers[t].start_pc);
			    if(lpass) fprintf(output,"$end_pc=%d;",handlers[t].end_pc);
			    if(lpass) fprintf(output,"$handler_pc=%d;",handlers[t].handler_pc);
			    if(lpass) fprintf(output,"$catch_type_index=%d;",handlers[t].catch_type_index);
			    if(lpass) if(handlers[t].start_pc==pos) printout("try_head");
			    if(lpass) if(handlers[t].end_pc==pos) printout("try_tail");
			    if(lpass) if(handlers[t].handler_pc==pos) printout("catch");
			}
		    }
		}


		if(casetags[pos])
		{
			if(lpass) fprintf(output,"$labelnr=%d;",labels[pos]);
			if(lpass) fprintf(output,"$case=%d;",cases[pos]);
			if(casetags[pos]==0) if(lpass) printout("clean");
			if(casetags[pos]==1) if(lpass) printout("case");
			if(casetags[pos]==2) if(lpass) printout("default");
		}
		else
		{
			if(labels[pos])
			{
				if(lpass) fprintf(output,"$labelnr=%d;",labels[pos]);
				if(lpass) printout("label");
			}
			else
			{
				if(lpass) printout("clean");
			}
		}

		u8 op=mem[pos++];
		if(!ops[op])
		{
			if(lpass) fprintf(stderr,"[%02x] at %d\n",op,fpos);
			error("ILLEGAL OPCODE\n");
		}
		char form2[8];
	    strcpy(form2,format[op]);
		if(op==0xc4) //WIDE
		{
			op=mem[pos++];
			strcpy(form2,format[op]);
			if(!iswide[op] || (format[op][0]!='b'  && format[op][0]!='c'))
			{
				if(lpass) fprintf(stderr,"illegal wide opcode [%02x]\n",op);
			}
			form2[0]='w';
		}
		if(lpass) fprintf(output,"$op=0x%02x;$name=\"%s\";",op,ops[op]);
		if(lpass) printout("op");
		u16 argn=0;
		char*form=form2;
		while(*form)
		{
			int i=0;
			switch(*form)
			{
				case 'b': 
					i=mem[pos++];
					if(lpass) fprintf(output,"$arg[%d]=%d;",argn++,i);
					break;
					
				case 'j':
				{
					s16 i;
					i=get2(&mem[pos]); 
					u32 jmpadr=oppos[nr]+i;
					pos+=2;
					if(lpass) fprintf(output,"$arg[%d]=%d;",argn++,jmpadr);
					int t,s=0;
					for(t=0;t<numorders;t++)
					{
						if(oppos[t]==jmpadr)
						{
							s=t;
						}
					};
					if(lpass) fprintf(output,"$jump=%d;",s-nr);
					if(lpass) fprintf(output,"$label=%d;",labels[jmpadr]);
					break;
				}
						
				case 'J':
				{
					u32 jmpadr=get4(&mem[pos]); pos+=4;
					fprintf(output,"$arg[%d]=%d;",argn++,jmpadr);
					int t,s=0;
					for(t=0;t<numorders;t++)
					{
						if(oppos[t]==jmpadr)
						{
							s=t;
						}
					};
					if(lpass) fprintf(output,"$jump=%d;",s-nr);
					if(lpass) fprintf(output,"$label=%d;",labels[jmpadr]);
					break;
				}
					
				case 'w': 
					i=get2(&mem[pos]); pos+=2;
					if(lpass) fprintf(output,"$arg[%d]=%d;",argn++,i);
					break;
				case 'c': 
					i=mem[pos++];
					if(lpass) fprintf(output,"$arg[%d]=%d;",argn++,i);
					break;
				case 's': 
					i=get2(&mem[pos]); pos+=2;
					if(lpass) fprintf(output,"$arg[%d]=%d;",argn++,i);
					break;
			    case '!':
					if(op==0xaa) //tableswitch
					{
						while(pos&3) pos++; // padding
						i=get4(&mem[pos]);pos+=4;
						if(lpass) fprintf(output,"$default=%d;",i);
						int low=get4(&mem[pos]);pos+=4;
						if(lpass) fprintf(output,"$low=%d;",low);
						int high=get4(&mem[pos]);pos+=4;
						if(lpass) fprintf(output,"$high=%d;",high);
						if(lpass) fprintf(output,"$num=%d;",high-low+1);
						int t;
						for(t=low;t<=high;t++)
						{
							i=get4(&mem[pos]);pos+=4;
							fprintf(output,"$key[%d]=%d;$offset[%d]=%d;",t-low,t,t-low,i);
						}
					}
					if(op==0xab) //lookupswitch
					{
						while(pos&3) pos++;
						i=get4(&mem[pos]);pos+=4;
						if(lpass) fprintf(output,"$default=%d;",i);
						int n=get4(&mem[pos]);pos+=4;
						if(lpass) fprintf(output,"$num=%d;",n);
						int t;
						for(t=0;t<n;t++)
						{
							int key=get4(&mem[pos]);pos+=4;
							int offset=get4(&mem[pos]);pos+=4;
							if(lpass) fprintf(output,"$key[%d]=%d;$offset[%d]=%d;",t,key,t,offset);
						}
					}
					break;
				    
			}
			form++;
		}
		if(lpass) printout(ops[op]);
		nr++;
	}
	if(lpass) printout("ops_tail");
	if(lpass) printout("method_tail");

	freecode();
}

/**
 * int onedescriptor(char*str)
 *
 * Converts a descriptor from the java from (e.g. [[LObject;
 * to a corresponding perl form.
 *
 * str		the descriptor to convert
 *
 */
int onedescriptor(char*str)
{
	char*old=str;
	int dim=0;
	int pos=0;
	if(str[pos]==',') pos++;
	while(str[pos]=='[')
	{
		pos++;
		dim++;
	}
	fprintf(output,"$dim=%d;",dim);
	if(strchr("bcdfijszvBCDFIJSZV",str[pos]))
	{
		fprintf(output,"$arg=\"%c\";",str[pos]);
		pos++;
	}
	else
	{
		if(str[pos]=='L')
		{
			pos++;
			fprintf(output,"$arg=\"");
			while(str[pos]!=';')
			{
				fprintf(output,"%c",str[pos]);
				pos++;
			}
			pos++;
			fprintf(output,"\";");
		}
		else
		{
			fprintf(stderr,"descriptor parse error \"%s\" (%s)\n",&str[pos],old);
			pos++;
		}
	}

	char a=str[pos];str[pos]=0;
	fprintf(output,"$descriptor=\"%s\";",str);
	str[pos]=a;

	return pos;
}

/**
 * int splitdescriptor(char*str)
 *
 * Converts a functiondescriptor from the java form (e.g. ([[LObject,I)[I;
 * to a corresponding perl form.
 *
 * str		the descriptor to convert
 *
 */
void splitdescriptor(char*str)
{
	int pos=1;
	int t=0;

	while(str[pos]!=')' && str[pos])
	{
		pos+=onedescriptor(&str[pos]);
		fprintf(output,"$descriptor[%d]=$descriptor;",t);
		fprintf(output,"$arg[%d]=$arg;",t);
		fprintf(output,"$dim[%d]=$dim;",t);
		t++;
	}
	
	fprintf(output,"$argnum=%d;",t);
	if(str[pos] && str[pos+1])
	{
		pos++;
		onedescriptor(&str[pos]);
		fprintf(output,"$returndescriptor=$descriptor;",t);
		fprintf(output,"$return=$arg;",t);
	    fprintf(output,"$returndim=$dim;",t);
	}
}

/**
 * void readmethods(int num)
 *
 * reads all methods of the class and performs the desired output
 *
 * num		the number of methods to read
 *
 */
void readmethods(int num)
{
	int t;
	for(t=0;t<num;t++)
	{
		if(t!=0) 
			printout("methods_separator");

		struct
		{
			u16 access_flags;
			u16 name_index;
			u16 type_index;
			u16 attributes_count;
		} head;

		input2(&head.access_flags);
		input2(&head.name_index);
		input2(&head.type_index);
		input2(&head.attributes_count);

		fprintf(output,"$name_index=%d;",head.name_index);
		fprintf(output,"$type_index=%d;",head.type_index);
		fprintf(output,"$attributes_count=%d;",head.attributes_count);

		char*name=&datapool[constantdatas[head.name_index]];
		char*type=&datapool[constantdatas[head.type_index]];

		fprintf(output,"$name=\"%s\";",formatstr(name));

		fprintf(output,"$access_flags=$flags=%d;"
				      "$ACC_PUBLIC=%d;"
			          "$ACC_FINAL=%d;"
			          "$ACC_PROTECTED=%d;"
			          "$ACC_STATIC=%d;"
			          "$ACC_FINAL=%d;"
			          "$ACC_SYNCHRONIZED=%d;"
			          "$ACC_NATIVE=%d;"
			          "$ACC_ABSTRACT=%d;"
					      ,head.access_flags,
					       (head.access_flags>>0)&1,
						   (head.access_flags>>1)&1,
						   (head.access_flags>>2)&1,
						   (head.access_flags>>3)&1,
						   (head.access_flags>>4)&1,
						   (head.access_flags>>5)&1,
						   (head.access_flags>>8)&1,
						   (head.access_flags>>10)&1);

		/*
		grab everything after the first ), i.e.
		(V)I becomes I
		[do we need this? would a splitdescriptor suffice?]
		*/
		char*ty=type;
		char*returnstr;
		int l=strlen(type);
		while(ty[l]!=')' && l>=0) l--;
		returnstr=&ty[l+1];
	
	    /**TAG:method_head: $return 
		**/
		fprintf(output,"$return=\"%s\";",returnstr);

		printout("method_head");
		
		splitdescriptor(type);
		printout("arguments_head");
		fprintf(output,"for($t=0;$t<$argnum;$t++) { $descriptor=$descriptor[$t];$arg=$arg[$t];$dim=$dim[$t]; if($t) { ");
		printout("arguments_separator");
		fprintf(output,"}");
		printout("argument");
		fprintf(output,"}");
		printout("arguments_tail");

		locals_count=0;
		fprintf(output,"$locals_count=0;");
		int s;
		long filepos=ftell(fi);
		int r;
	
		handlers=0;
		handlers_count=0;
		for(lpass=0;lpass<2;lpass++)
		{
			if(lpass) fseek(fi,filepos,SEEK_SET);
			for(s=0;s<head.attributes_count;s++)
			{
				struct
				{
					u16 attribute_name_index;
					u32 bytes_count;
				} head2;

				input2(&head2.attribute_name_index);
				input4(&head2.bytes_count);

				char*data=new char[head2.bytes_count];
				input(data,head2.bytes_count);

				char*str=&datapool[constantdatas[head2.attribute_name_index]];

				if(!strcmp("Exceptions",str))
				{
					u16 throws_count=get2(&data[0]);
					//reverse2(&throws_count);

					if(lpass) fprintf(output,"$i=%d;$num=%d;",s,throws_count);
					if(lpass) printout("throws_head");

					int t;
					for(t=0;t<throws_count;t++)
					{
						if(lpass) if(!t) printout("throws_separator");
						u16 mythrow;
						mythrow=get2(&data[t*2+2]);

						if(lpass) fprintf(output,"$i=%d;$index=%d;",t,mythrow);
						if(lpass) fprintf(output,"$name=$constantpool1[$index];");

						if(lpass) printout("throw");
					}
					if(lpass) printout("throws_tail");
				}
				else
				if(!strcmp("Code",str))
				{
					struct
					{
						u16 max_stack;
						u16 max_locals;
						u32 code_count;
					} head3;
					mem=&data[0];
					
					inputmem2(&head3.max_stack);
					inputmem2(&head3.max_locals);
					inputmem4(&head3.code_count);

					if(lpass) fprintf(output,"$name=\"%s\";$max_stack=%d;$max_locals=%d;$code_count=%d;",formatstr(name),head3.max_stack,head3.max_locals,head3.code_count);

					if(lpass) 
					{	readcode(lpass,(u8*)mem,head3.code_count);
						int t;
						for(t=0;t<locals_count;t++)
						{
							fprintf(output,"$localname[%d]=",t);
						}
						if(t>0)
						fprintf(output,"\"\";\n");
					}
					mem+=head3.code_count;
		
					inputmem2(&handlers_count);
					if(!lpass)
					{
						handlers=new handler_type[handlers_count];
					}

					int t;
					for(t=0;t<handlers_count;t++)
					{
						
						handler_type head4;
						inputmem2(&head4.start_pc);
						inputmem2(&head4.end_pc);
						inputmem2(&head4.handler_pc);
						inputmem2(&head4.catch_type_index);

						handlers[t]=head4;
					}

					u16 attributes_count;
					inputmem2(&attributes_count);

					for(t=0;t<attributes_count;t++)
					{
						struct
						{
							u16 attribute_name_index;
							u32 bytes_count;
						} head4;
						inputmem2(&head4.attribute_name_index);
						inputmem4(&head4.bytes_count);
						
						char*data2=mem;mem+=head4.bytes_count;

						char*name2=&datapool[constantdatas[head4.attribute_name_index]];
		
						if(!strcmp("LineNumberTable",name2))
						{
							if(lpass) printout("linenumbers_head");
							char*memold=mem;
							mem=data2;
							inputmem2(&lines_count);
							int t;
							for(t=0;t<lines_count;t++)
							{
								if(lpass) if(t) printout("linenumbers_seperator");
								inputmem2(&start_pc[t]);
								inputmem2(&line_number[t]);
								if(t>0 && start_pc[t-1]>start_pc[t])
								{
									fprintf(stderr,"Broken Line Number Table\n");
									lines_count=0;break;
								}

								if(lpass) fprintf(output,"$start_pc=%d;$line_number=%d;",start_pc[t],line_number[t]);
								if(lpass) printout("linenumber");
							}
							start_pc[lines_count]=0xffff;
							if(lpass) printout("linenumbers_tail");

							mem=memold;
						}
						else
						if(!strcmp("LocalVariableTable",name2))
						{
							char*memold=mem;
							mem=data2;
							u16 local_variables_count;
							inputmem2(&local_variables_count);
							locals_count=local_variables_count;
							fprintf(output,"$locals_count=%d+0;",locals_count);
							if(lpass) printout("localvars_head");
							int t;
							for(t=0;t<local_variables_count;t++)
							{
								if(t)
								{
									if(lpass) printout("localvars_seperator");
								}

								localvar loc;

								inputmem2(&loc.start_pc);
								inputmem2(&loc.length);
								inputmem2(&loc.name_index);
								inputmem2(&loc.type_index);
								inputmem2(&loc.slot);

								local_vars[t]=loc;

								if(lpass) fprintf(output,"$start_pc=%d;",loc.start_pc);
								if(lpass) fprintf(output,"$length=%d;",loc.length);
								if(lpass) fprintf(output,"$name_index=%d;",loc.name_index);
								if(lpass) fprintf(output,"$type_index=%d;",loc.type_index);
								if(lpass) fprintf(output,"$slot=%d;\n",loc.slot);

								if(lpass) printout("localvar");
							}
							mem=memold;
							if(lpass) printout("localvars_tail");
						}
						else
						{
							if(lpass) fprintf(stderr,"Unrecognized Codeattribute: %s (%d bytes)\n",name2,head4.bytes_count);
						}
					}
				}
				else
				{
					if(lpass) fprintf(stderr,"Unrecognized Methodattribute: %s (%d bytes)\n",str,head2.bytes_count);
				}

				delete data;
			}
		}
	
		if(handlers)
		delete handlers;

		/*
		start printing as usual again
		*/
		lpass=1;
	}
	printout("methods_tail");
}

/**
 * void storeinfo(char*name,int val);
 *
 * sets a perl variable to a specific value
 *
 */
void storeinfo(char*name,int val)
{
	fprintf(output,"$%s=%d;",name,val);
}

/**
 * void storeinfo(char*name,char*string);
 *
 * sets a perl variable to a specific string
 *
 */
void storeinfo(char*name,char*val)
{
	fprintf(output,"$%s=\"%s\";",name,val);
}

/**
 * void readclass()
 *
 * reads the classfile. Called from main()
 *
 */
void readclass()
{
	struct head 
	{
		u32 magic;
		u16 minor_version;
		u16 major_version;
		u16 pool_count;
	} head;
	
	input4(&head.magic);
	input2(&head.minor_version);
	input2(&head.major_version);
	input2(&head.pool_count);

	if(head.magic!=0xCAFEBABE) 
		error("No class file.\n");

	fprintf(output,"$filename=$name=\"%s\";",fname);
	fprintf(output,"$major=%d;$minor=%d;",head.major_version,
		                              head.minor_version);
	printout("class_head");

	fprintf(output,"$pool_count=$num=%d;",head.pool_count);
	printout("constants_head");

	readconstantpool(head.pool_count);

	struct head2
	{
		u16 access_flags;
		u16 this_class;
		u16 super_class;
		u16 interfaces_count;

	} head2;

	input2(&head2.access_flags);
	input2(&head2.this_class);
	input2(&head2.super_class);
	input2(&head2.interfaces_count);

	fprintf(output,"$class_flags=$flags=%d;$ACC_PUBLIC=%d;"
			          "$ACC_FINAL=%d;"
					  "$ACC_SUPER=%d;"
					  "$ACC_INTERFACE=%d;"
					  "$ACC_ABSTRACT=%d;"
					      ,head2.access_flags,
					       head2.access_flags&1,
						   (head2.access_flags>>2)&1,
						   (head2.access_flags>>5)&1,
						   (head2.access_flags>>9)&1,
						   (head2.access_flags>>10)&1);
	printout("access_flags");
	
	fprintf(output,"$i=%d;$this_name=$name=$constantpool1[$constantpool1[$i]];",head2.this_class);
	printout("this_class");

	fprintf(output,"$i=%d;$super_name=$name=$pool[$pool[$i]];",head2.super_class);
	printout("super_class");

	fprintf(output,"$interfaces_count=$num=%d;",head2.interfaces_count);
    printout("interfaces_head"); 
	readinterfaces(head2.interfaces_count);

    u16 fields_count;
	input(&fields_count,2);
	reverse2(&fields_count);

	fprintf(output,"$fields_count=$num=%d;",fields_count);
    printout("fields_head");
	readfields(fields_count);

	u16 methods_count;
	input(&methods_count,2);
	reverse2(&methods_count);

	fprintf(output,"$methods_count=$num=%d;",methods_count);
    printout("methods_head");
	readmethods(methods_count);
	
	u16 attributes_count;
	input(&attributes_count,2);
	reverse2(&attributes_count);

	fprintf(output,"$attributes_count=$num=%d;",attributes_count);
    printout("attributes_head");
	readattributes(attributes_count);

}

/**
 * printusage(char*name)
 *
 * prints out the usage of the program
 */
void printusage(char*name)
{
	fprintf(stderr,"Usage: %s [flags] inputfile [formatfile [outputfile]]\n",name);
	fprintf(stderr,"\n");
	fprintf(stderr,"flags:\n");
	fprintf(stderr,"-v             : Be verbose\n");
	fprintf(stderr,"-i includepath : Path's for include files, seperated by ';'\n");
	fprintf(stderr,"-h             : This page\n");
	fprintf(stderr,"-f name        : Formatfile\n");
	fprintf(stderr,"-o name        : Name of output file\n");
	exit(1);
}
void printhelpconfig()
{
	fprintf(stderr,"for jissa to disassemble anything, there has to\n");
	fprintf(stderr,"be at least one config file. these files describe \n");
	fprintf(stderr,"how the output files looks like. In the unpacked jissa \n");
	fprintf(stderr,"distribution, the config files are located under ./lib \n");
	fprintf(stderr,"you'll have to copy them somewhere where jissa can find them, \n");
	fprintf(stderr,"preferably /etc/jissa. (call jissa with the -v option in order\n");
	fprintf(stderr,"to gain some knowledge about the order in which jissa tries to\n");
	fprintf(stderr,"find the config files)\n");
	exit(1);
}

char*destname=0;
char*fmtname=0;
char strbuf2[256];

char*argvx[20];
char*argv[20];
int argn;

void getflags(int argn2,char*argv2[])
{
	int t;argn=0;
	for(t=0;t<argn2;t++)
	{
		if(argv2[t][0]=='-')
		{
			switch(argv2[t][1])
			{
				case 'h':
				case '-': 
				{
				    if(!strcmp(argv2[t],"--help-config"))
					printhelpconfig();
				    else
					printusage(argv2[0]);
				}
				case 'v': verbose=1;break;

				case 'i': 
					if(t==argn2-1)
					{
						fprintf(stderr,"-i requires an argument!\n\n");
						printusage(argv2[0]);
					}
					includefiles_append(argv2[++t]);
				break;
				case 'o': 
					if(t==argn2-1)
					{
						fprintf(stderr,"-o requires an argument!\n\n");
						printusage(argv2[0]);
					}
					destname=argv2[++t];
				break;
				case 'f':
					if(t==argn2-1)
					{
						fprintf(stderr,"-f requires an argument!\n\n");
						printusage(argv2[0]);
					}
					fmtname=argv2[++t];
				break;
			}
		}
		else
			argv[argn++]=argv2[t];
	}
}
void readflash();

void main(int argn2,char*argv2[])
{
	getflags(argn2,argv2);
	includefiles_append("/etc/jissa/");
	includefiles_append("./");

	strcpy(basepath,argv[0]);
	int l=strlen(basepath);
	while(basepath[l]!='/' && l>0) l--;
	if(l==0)
	{
		basepath[0]='.';l=1;
	}
	basepath[l]='/';
	basepath[l+1]=0;
	includefiles_append(basepath);

	char fmtisstandard=0;
	switch(argn)
	{
		case 2: fname=argv[1];
				if(!fmtname)
				{
				    fmtname="standard";
				    fmtisstandard=1;
				}
				destname="stdout";
				break;
		case 3:	fname=argv[1];
				if(!destname)
				{
				    destname="stdout";
				}
				fmtname=argv[2];
				break;
		case 4:	fname=argv[1];
				fmtname=argv[2];
				destname=argv[3];
				break;
		default:
				printusage(argv[0]);
	}

	if(verbose)
	fprintf(stderr,"%s => %s => %s\n",fname,fmtname,destname);

	detectreverse();
	mkops();

        char java=0;
	char flash=0;
	u8 a;
	fi=fopen(fname,"rb");
	fread(&a,1,1,fi);
	if(a==0xCA) java=1;
	if(a==0x46) flash=1;
	fclose(fi);

	if((java|flash)==0) 
	{
	    fprintf(stderr,"Unknown file format (neither flash nor java): %s\n",fname);
	    return 1;
	}

	if(flash==1 && fmtisstandard)
	    fmtname="flash";

	char*tempname="/tmp/output.pl";
		
	output=fopen(tempname,"w");
	if(!output)
	{
		fprintf(stderr,"Couldn't create temporary file \'%s\'!\n",tempname);
		return 1;
	}

	if(verbose)
	fprintf(stderr,"Using Format \"%s\"\n",fmtname);
	readconfig(fmtname);
	
	char systxt[160];

	if(strcmp(destname,"stdout")) 
	{
		sprintf(systxt,"perl %s > %s",tempname,destname);
	}
	else				
	{		
		sprintf(systxt,"perl %s",tempname);
	}

	pass=0;
	fi=fopen(fname,"rb");fpos=0;
	if(!fi) {fprintf(stderr,"File not found: %s\n",fname);return 0;}
	if(java)
	    readclass();
	else
	    readflash();
	fclose(fi);
	pass=1;
	fi=fopen(fname,"rb");fpos=0;
	if(!fi) {fprintf(stderr,"File not found: %s\n",fname);return 0;}
	if(java)
	    readclass();
	else
	    readflash();
	fclose(fi);

	system(systxt);
}

#include "flash.c++"

