#include "flash.h"

u8 bitpos=8,mybyte;
void resetbits()
{
    bitpos=8;
}
u8 bitmem=0;
u32 readbit()
{
    if(bitpos==8) 
    {
	bitpos=0;
	if(bitmem)
	    inputmem1(&mybyte);
	else
	    input1(&mybyte);
    }
    return (mybyte>>(7-bitpos++))&1;
}
void readbits(u32*val,int num)
{
    int t;
    *val=0;
    for(t=0;t<num;t++)
    {
	*val<<=1;
	*val|=readbit();
    }
}

RGB readRGB()
{
    RGB rgb;
    inputmem1(&rgb.r);
    inputmem1(&rgb.g);
    inputmem1(&rgb.b);
    return rgb;
}

char* readString()
{
    char*now=mem;
    char a;
    do
    {
	inputmem1(&a);
    }
    while(a);
    return now;
}

u16 readu16()
{
    u8 a,b;
    // I'm not using inputmem2(&a) here because our input is
    // little endian.
    inputmem1(&a);
    inputmem1(&b);
    return ((u16)b)*256+a;
}

RECT readRECT()
{
    u32 a,b;
    RECT r;
    bitmem=1;
    readbits(&a,5);
    readbits(&b,a);
    r.x1=b;
    readbits(&b,a);
    r.x2=b;
    readbits(&b,a);
    r.y1=b;
    readbits(&b,a);
    r.y2=b;
    bitmem=0;
    return r;
}

void readflashdata(int pos, char type)
{
    switch(type)
    {
	case 'c':
	    RGB c=readRGB();
	    fprintf(output,"$col%d_r=%d;\n",pos,c.r);
	    fprintf(output,"$col%d_g=%d;\n",pos,c.g);
	    fprintf(output,"$col%d_b=%d;\n",pos,c.b);
	break;
	case 's':
	{
	    char*st=readString();
	    fprintf(output,"$str%d=\"%s\";",st);
	}
	break;
	case 'r':
	{
	    RECT r=readRECT();
	    fprintf(output,"$b%d_x1=%d;",pos,r.x1);
	    fprintf(output,"$b%d_y1=%d;",pos,r.y1);
	    fprintf(output,"$b%d_x2=%d;",pos,r.x2);
	    fprintf(output,"$b%d_y2=%d;",pos,r.y2);
	}
	break;
	case 'i':
	case '6':
	    u16 a=readu16();
	    fprintf(output,"$word%d=%d;",pos,a);
	    if(type=='i')
		fprintf(output,"$id=%d;",a);
	break;
    }
}

struct numname
{
    int num;
    char*name;
    char*datatype;
    numname(int num,char*name,char*datatype) :num(num),name(name), 
	datatype(datatype) {}
}
idtoname[]=
{
    //controltags
    numname(9,"backgroundcolor","c"),
    numname(43,"framelabel","s"),
    numname(24,"protect","-"),
    numname(0,"end","-"),
    numname(56,"exportassets","66s6s"),
    numname(57,"importassets","s66s6s"),
    numname(58,"enabledebugger","s"),
    numname(1023,"definebitsptr","3"),
    //actiontags
    numname(12,"doaction","A"),
    /*
    numname(0x81,"ActionGotoFrame"),
    numname(0x83,"ActionGetURL"),
    numname(0x04,"ActionNextFrame"),
    numname(0x05,"ActionPrevFrame"),
    numname(0x06,"ActionPlay"),
    numname(0x07,"ActionStop"),
    numname(0x08,"ActionToggleQuality"),
    numname(0x09,"ActionStopSounds"),
    numname(0x8a,"ActionWaitForFrame"),
    numname(0x8b,"ActionSetTarget"),
    numname(0x8c,"ActionGotoLabel"),
     */

    numname(2,"defineshape","ir"),
    numname(22,"defineshape2","ir"),
    numname(32,"defineshape3","ir"),

    numname(46,"definemorphshape","irr"),

    numname(4,"placeobject","i6"),
    numname(26,"placeobject2",""),
    numname(5,"removeobject","i6"),
    numname(28,"removeobject2","6"),
    numname(1,"showframe",""),
    numname(10,"definefont","i"),    
    numname(13,"definefontinfo","i"),
    numname(48,"definefont2","i"),
    numname(11,"definetext","i"),
    numname(33,"definetext2","i"),
    numname(37,"defineedittext","i"),

    numname(14,"definesound",""),
    numname(15,"startsound",""),
    numname(18,"soundstreamhead",""),
    numname(45,"soundstreamhead2",""),
    numname(19,"soundstreamblock",""),

    numname(39,"definesprite",""),
    numname(38,"definemovie",""),

    numname(6,"definebits","i"),
    numname(21,"definebitsjpeg2","i"),
    numname(35,"definebitsjpeg3","i"),
    numname(20,"definebitslossless","i"),
    numname(20,"definebitslossless2","i"),
    numname(8,"jpegtables",""),
    numname(7,"definebutton",""),
    numname(34,"definebutton2",""),
    numname(23,"definebuttoncxform",""),
    numname(17,"definebuttonsound","")
};

int idtonamelen=sizeof(idtoname)/sizeof(numname);


void readRECT(char*name)
{
    u32 a,b;
    readbits(&a,5);

    readbits(&b,a);
    fprintf(output,"$%s_x1=%d;\n",name,b);
    
    readbits(&b,a);
    fprintf(output,"$%s_x2=%d;\n",name,b);

    readbits(&b,a);
    fprintf(output,"$%s_y1=%d;\n",name,b);

    readbits(&b,a);
    fprintf(output,"$%s_y2=%d;\n",name,b);
}

void readflash()
{
    if(pass==0) 
	doreverse^=1;

    char a[4],version;
    u32 length;
    input1(&a[0]);
    input1(&a[1]);
    input1(&a[2]);
    a[3]=0;
    if(strcmp(a,"FWS"))
    {
	printf("Not a flash file!\n");
    }
    input1(&version);
    storeinfo("version",(int)version);
    input4(&length);
    storeinfo("length",length);
    resetbits();
    readRECT("framesize");
    u16 rate;
    input2(&rate);
    fprintf(output,"$framerate=$rate=%f;",(rate>>8)+((rate&255)/256.0));
    u16 count;
    input2(&count);
    storeinfo("framecount",(int)count);
    printout("flash_head");

    while(1)
    {
	u16 tag;
	input2(&tag);
	u32 taglength=tag&0x3f;
	u32 tagid=tag>>6;
	if(taglength==0x3f)
	{
	    input4(&taglength);
	}
	int t;
	u8*data=new u8[taglength];
	if(!data)
	{
	    fprintf(stderr,"Out of memory. Invalid tag.\n");
	    data=0;
	} 
	else
	{
	    for(t=0;t<taglength;t++)
	    {
		u8 a;
		input1(&data[t]);
	    }
	    mem=(char*)data;
	}
	fprintf(output,"#TAG%03x %d bytes\n",tagid,taglength);
	fprintf(output,"$tag=0x%03x;$taglength=%d;\n",tagid,taglength);
	int found=0;
	for(t=0;t<idtonamelen;t++) // should be done by a class, once
	    //and for all
	{
	    if(tagid==idtoname[t].num)
	    {
		fprintf(output,"$name=\"%s\";\n",idtoname[t].name);
		found=1;
		int s=0;
		while(idtoname[t].datatype[s])
		{
		    readflashdata(s+1,idtoname[t].datatype[s]);
		    s++;
		}
		printout("tag");
		printout(idtoname[t].name);
	    }
	}

	if(data)
	    delete data;

	if(!found)
	{
	    printout("unknowntag");
	}
	if(tagid==0) break; //END TAG
    }
    printout("flash_tail");
}

