/* JISSA- Generic Disassembler for Java-Bytecode
 
   ops.c++
   opcode definitions

   Copyright (c) 2000 Matthias Kramm <kramm@iname.com> 

   This file is distributed under the GPL, see file COPYING for details */



#include "stdlib.h"
#include "stdio.h"

char*ops[256];
int stack[256];
char*format[256];
char iswide[256];

/**
 * void isop(char*name,int mstack,int op,char*mformat)
 *
 * Create an opcode
 *
 * name		the name of the opcode
 * mstack	n2-n1, where n1 is the stack position before and n1 is the
 *			stack position after execution of the opcode
 * op		opcode
 * mformat	description of the data typically following the opcode
 *			b=byte,w=word,d=dword,j=jump address,J=big jump address
 *			t and l are special and used for table and lookup-switches
 *
 */
void isop(char*name,int mstack,int op,char*mformat)
{
	if(ops[op])
		printf("DOUBLE: %02x\n",op);
	ops[op]=name;
	int l=strlen(name);
	iswide[op]=0;
	if(l>=2)
	if(name[l-1]=='w' && name[l-2]=='/')
	{
		char*a=new char[l+1];strcpy(a,name);
		a[l-2]=0;
		ops[op]=a;
		iswide[op]=1;
	}
	stack[op]=mstack;
	format[op]=mformat;
}

/**
 * void mkops()
 *
 * Create all known legal opcodes
 *
 */
void mkops()
{
	int t;
	for(t=0;t<256;t++)
	{
		ops[t]=0;
		stack[t]=0;
	}

	isop("aaload",-1,0x32,"");
	isop("aastore",-3,0x53,"");
	isop("aconst_null",1,0x01,"");
	isop("aload/w",1,0x19,"b");
	isop("aload_0",1,0x2a,"");
	isop("aload_1",1,0x2b,"");
	isop("aload_2",1,0x2c,"");
	isop("aload_3",1,0x2d,"");
	isop("anewarray",0,0xbd,"w");
	isop("areturn",0,0xb0,"");
	isop("arraylength",0,0xbe,"");
	isop("astore/w",-1,0x3a,"b");
	isop("astore_0",-1,0x4b,"");
	isop("astore_1",-1,0x4c,"");
	isop("astore_2",-1,0x4d,"");
	isop("astore_3",-1,0x4e,"");
	isop("athrow",0,0xbf,"");
	isop("baload",-1,0x33,"");
	isop("bastore",-3,0x54,"");
	isop("bipush",1,0x10,"c");
	isop("breakpoint",0,0xca,"");
	isop("caload",-1,0x34,"");
	isop("castore",-3,0x55,"");
	isop("checkcast",0,0xc0,"w");
	isop("d2f",-1,0x90,"b");
	isop("d2i",-1,0x8e,"");
	isop("d2l",0,0x8f,"");
	isop("dadd",-2,0x63,"");
	isop("daload",0,0x31,"");
	isop("dastore",-4,0x52,"");
	isop("dcmpg",-3,0x98,"");
	isop("dcmpl",-3,0x97,"");
	isop("dconst_0",2,0x0e,"");
	isop("dconst_1",2,0x0f,"");
	isop("ddiv",-2,0x6f,"");
	isop("dload/w",2,0x18,"b");
	isop("dload_0",2,0x26,"");
	isop("dload_1",2,0x27,"");
	isop("dload_2",2,0x28,"");
	isop("dload_3",2,0x29,"");
	isop("dmul",-2,0x6b,"");
	isop("dneg",0,0x77,"");
	isop("drem",-2,0x73,"");
	isop("dreturn",0,0xaf,"");
	isop("dstore/w",-2,0x39,"b");
	isop("dstore_0",-2,0x47,"");
	isop("dstore_1",-2,0x48,"");
	isop("dstore_2",-2,0x49,"");
	isop("dstore_3",-2,0x4a,"");
	isop("dsub",-2,0x67,"");
	isop("dup",1,0x59,"");
	isop("dup2",2,0x5c,"");
	isop("dup2_x1",2,0x5d,"");
	isop("dup2_x2",2,0x5e,"");
	isop("dup_x1",1,0x5a,"");
	isop("dup_x2",1,0x5b,"");
	isop("f2d",1,0x8d,"");
	isop("f2i",0,0x8b,"");
	isop("f2l",1,0x8c,"");
	isop("fadd",-1,0x62,"");
	isop("faload",-1,0x30,"");
	isop("fastore",-3,0x51,"");
	isop("fcmpg",-1,0x96,"");
	isop("fcmpl",-1,0x95,"");
	isop("fconst_0",1,0x0b,"");
	isop("fconst_1",1,0x0c,"");
	isop("fconst_2",1,0x0d,"");
	isop("fdiv",-1,0x6e,"");
	isop("fload/w",1,0x17,"b");
	isop("fload_0",1,0x22,"");
	isop("fload_1",1,0x23,"");
	isop("fload_2",1,0x24,"");
	isop("fload_3",1,0x25,"");
	isop("fmul",-1,0x6a,"");
	isop("fneg",0,0x76,"");
	isop("frem",-1,0x72,"");
	isop("freturn",0,0xae,"");
	isop("fstore/w",-1,0x38,"b");
	isop("fstore_0",-1,0x43,"");
	isop("fstore_1",-1,0x44,"");
	isop("fstore_2",-1,0x45,"");
	isop("fstore_3",-1,0x46,"");
	isop("fsub",-1,0x66,"");
	isop("getfield",64,0xb4,"w");
	isop("getstatic",65,0xb2,"w");
    isop("goto",0,0xa7,"j");
	isop("goto_w",0,0xc8,"J");
	isop("i2b",0,0x91,"");
	isop("i2c",0,0x92,"");
	isop("i2d",1,0x87,"");
	isop("i2f",0,0x86,"");
	isop("i2l",1,0x85,"");
	isop("i2s",0,0x93,"");
	isop("iadd",-1,0x60,"");
	isop("iaload",-1,0x2e,"");
	isop("iand",-1,0x7e,"");
	isop("iastore",-3,0x4f,"");
	isop("iconst_m1",1,0x02,"");
	isop("iconst_0",1,0x03,"");
	isop("iconst_1",1,0x04,"");
	isop("iconst_2",1,0x05,"");
	isop("iconst_3",1,0x06,"");
	isop("iconst_4",1,0x07,"");
	isop("iconst_5",1,0x08,"");
	isop("idiv",-1,0x6c,"");
	isop("if_acmpeq",-2,0xa5,"j");
	isop("if_acmpne",-2,0xa6,"j");
	isop("if_icmpeq",-2,0x9f,"j");
	isop("if_icmpge",-2,0xa2,"j");
	isop("if_icmpgt",-2,0xa3,"j");
	isop("if_icmple",-2,0xa4,"j");
	isop("if_icmplt",-2,0xa1,"j");
	isop("if_icmpne",-2,0xa0,"j");
	isop("ifeq",-1,0x99,"j");
	isop("ifge",-1,0x9c,"j");
	isop("ifgt",-1,0x9d,"j");
	isop("ifle",-1,0x9e,"j");
	isop("iflt",-1,0x9b,"j");
	isop("ifne",-1,0x9a,"j");
	isop("ifnonnull",-1,0xc7,"j");
	isop("ifnull",-1,0xc6,"j");
	isop("iinc/w",0,0x84,"bc");
	isop("iload/w",1,0x15,"b");
	isop("iload_0",1,0x1a,"");
	isop("iload_1",1,0x1b,"");
	isop("iload_2",1,0x1c,"");
	isop("iload_3",1,0x1d,"");
	isop("impdep1",0,0xfe,"");
	isop("impdep2",0,0xff,"");
	isop("imul",-1,0x68,"");
	isop("ineg",0,0x74,"");
	isop("instanceof",0,0xc1,"w");
	isop("invokeinterface",32,0xb9,"wbb");
	isop("invokespecial",32,0xb7,"w");
	isop("invokestatic",31,0xb8,"w");
	isop("invokevirtual",32,0xb6,"w");
	isop("ior",-1,0x80,"");
	isop("irem",-1,0x70,"");
	isop("ireturn",0,0xac,"");
	isop("ishl",-1,0x78,"");
	isop("ishr",-1,0x7a,"");
	isop("istore/w",-1,0x36,"b");
	isop("istore_0",-1,0x3b,"");
	isop("istore_1",-1,0x3c,"");
	isop("istore_2",-1,0x3d,"");
	isop("istore_3",-1,0x3e,"");
	isop("isub",-1,0x64,"");
	isop("iushr",-1,0x7c,"");
	isop("ixor",-1,0x82,"");
	isop("jsr",1,0xa8,"j");
	isop("jsr_w",1,0xc9,"J");
	isop("l2d",0,0x8a,"");
	isop("l2f",-1,0x89,"");
	isop("l2i",-1,0x88,"");
	isop("ladd",-2,0x61,"");
	isop("laload",0,0x2f,"");
	isop("land",-2,0x7f,"");
	isop("lastore",-4,0x50,"");
	isop("lcmp",-3,0x94,"");
	isop("lconst_0",2,0x09,"");
	isop("lconst_1",2,0x0a,"");
	isop("ldc",1,0x12,"b");
	isop("ldc2_w",2,0x14,"w");
	isop("ldc_w",1,0x13,"w");
	isop("ldiv",-2,0x6d,"");
	isop("lload/w",2,0x16,"b");
	isop("lload_0",2,0x1e,"");
	isop("lload_1",2,0x1f,"");
	isop("lload_2",2,0x20,"");
	isop("lload_3",2,0x21,"");
	isop("lmul",-1,0x69,"");
	isop("lneg",0,0x75,"");
	isop("lookupswitch",-1,0xab,"!l");
	isop("lor",-2,0x81,"");
	isop("lrem",-2,0x71,"");
	isop("lreturn",0,0xad,"");
	isop("lshl",-1,0x79,"");
	isop("lshr",-1,0x7b,"");
	isop("lstore/w",-2,0x37,"b");
	isop("lstore_0",-2,0x3f,"");
	isop("lstore_1",-2,0x40,"");
	isop("lstore_2",-2,0x41,"");
	isop("lstore_3",-2,0x42,"");
	isop("lsub",-2,0x65,"");
	isop("lushr",-1,0x7d,"");
	isop("lxor",-2,0x83,"");
	isop("monitorenter",-1,0xc2,"");
	isop("monitorexit",-1,0xc3,"");
	isop("multianewarray",34,0xc5,"");
	isop("new",1,0xbb,"w");
	isop("newarray",0,0xbc,"b");
	isop("nop",0,0,"");
	isop("pop",-1,0x57,"");
	isop("pop2",-2,0x58,"");
	isop("putfield",-2,0xb5,"w");
	isop("putstatic",-2,0xb3,"w");
	isop("ret",0,0xa9,"w");
	isop("return",0,0xb1,"");
	isop("saload",-1,0x35,"");
	isop("sastore",-3,0x56,"");
	isop("sipush",1,0x11,"s");
	isop("swap",0,0x5f,"");
	isop("tableswitch",-1,0xaa,"!t");
	isop("WIDE",0,0xc4,"");
}

