/* JISSA- Generic Disassembler for Java-Bytecode
 
 split.c++
 routines for datatype conversion

 Copyright (c) 2000 Matthias Kramm <kramm@iname.com> 

 This file is distributed under the GPL, see file COPYING for details */


#include "stdlib.h"
#include "stdio.h"

char doreverse=0;

/*
 * datatype definitions
 */
typedef unsigned long long int u64;
typedef unsigned long int u32;
typedef unsigned short int u16;
typedef unsigned char u8;
typedef signed long long int s64;
typedef signed long int s32;
typedef signed short int s16;
typedef signed char s8;

/**
 * char splitbuf[]
 *
 * buffer for string operations
 *
 */
char splitbuf[65536*2];

/**
 * void reverse2(void*a)
 *
 * reverse the word at a given memory location. the goal of this is to bring
 * all data from low endian to the current machine's format. This is why we
 * test doreverse, which is 1 if the current machine is low endian
 *
 * a		the memory location where the inversion should be performed
 *
 */
void reverse2(void*a)
{
	if(doreverse)
	{
		char i=((char*)a)[0];
		char j=((char*)a)[1];
		((char*)a)[1]=i;
		((char*)a)[0]=j;
	}
}

/**
 * void reverse4(void*a)
 *
 * like reverse2, just for doublewords
 *
 * a		the memory location where the inversion should be performed
 *
 */
void reverse4(void*a)
{
	if(doreverse)
	{
		char i=((char*)a)[0];
		char j=((char*)a)[1];
		char k=((char*)a)[2];
		char l=((char*)a)[3];
		((char*)a)[3]=i;
		((char*)a)[2]=j;
		((char*)a)[1]=k;
		((char*)a)[0]=l;
	}
}

/*
 * void reverse8(void*a)
 *
 * like reverse2 just for quadwords
 *
 * a		the memory location where the inversion should be performed
 *
*/
void reverse8(void*a)
{
	if(doreverse)
	{
		char i=((char*)a)[0];
		char j=((char*)a)[1];
		char k=((char*)a)[2];
		char l=((char*)a)[3];
		char m=((char*)a)[4];
		char n=((char*)a)[5];
		char o=((char*)a)[6];
		char p=((char*)a)[7];
		((char*)a)[7]=i;
		((char*)a)[6]=j;
		((char*)a)[5]=k;
		((char*)a)[4]=l;
		((char*)a)[3]=m;
		((char*)a)[2]=n;
		((char*)a)[1]=o;
		((char*)a)[0]=p;
	}
}

/**
 * void detectreverse()
 *
 *  detect machine byte ordering
 * doreverse==0: machine operates with network byte order (big endian)
 * doreverse==1: machine operates with intel byte order (little endian)
 *
 * (doreverse==0 means: we don't need to reverse the bytes read from the 
 *                      classfile (as these are already network byte order)
*/
void detectreverse()

{
	// BEBAFECA?
	int a=1;
	doreverse=(*(char*)&a);
}

/**
 * char* formatstr(char*str)
 *
 * Convert special chars in a string to their escaped form. E.g.
 * " $ " gets " \$ " Also converts nonprintable characters to _
 *
 * str	the string to convert
 * returns	the converted string
 */
char* formatstr(char*str)
{
	int fbpos=0;
	while(*str)	
	{
		switch(*str)
		{
			case '@': 
			splitbuf[fbpos++]='\\';splitbuf[fbpos++]='$';break;
			case '$': 
			splitbuf[fbpos++]='\\';splitbuf[fbpos++]='$';break;
			case '\n': 
			splitbuf[fbpos++]='\\';splitbuf[fbpos++]='n';break;
			case '\t': 
			splitbuf[fbpos++]='\\';splitbuf[fbpos++]='t';break;
			case '\\':
			splitbuf[fbpos++]='\\';splitbuf[fbpos++]='\\';break;
			case '"':
			splitbuf[fbpos++]='\\';splitbuf[fbpos++]='"';break;
			case '%':
			splitbuf[fbpos++]='\\';splitbuf[fbpos++]='%';break;
			default:
			splitbuf[fbpos++]=((*str&128) || (*str<32))?'_':*str;
		}
		str++;
	}
	splitbuf[fbpos]=0;
	return splitbuf;
}

void eight2four(u64 a,u32*a1,u32*a2)
{
	if(doreverse)
	{
		*a1=((u32*)&a)[0];
		*a2=((u32*)&a)[1];
	}
	else
	{
		*a1=((u32*)&a)[1];
		*a2=((u32*)&a)[0];
	}
}

void four2eight(u64*a,u32 a1,u32 a2)
{
	if(doreverse)
	{
		((u32*)a)[0]=a1;
		((u32*)a)[1]=a2;
	}
	else
	{
		((u32*)a)[1]=a2;
		((u32*)a)[0]=a1;
	}
}

/** 
 * void splitlong(u64 val, char** str)
 *
 * converts a quadword into a string
 *
 * val		the quadword to convert
 * str		a pointerreference to the destination string
 *
 */
void splitlong(u64 val, char**str)
{
	u32 val1,val2;
	splitbuf[255]=0;
	char*pos=&splitbuf[254];
    while(val!=0)
    {
		*pos=0x30+(val%10);
		pos--;
		val=val/10;
	}
	*str=&pos[1];
}

/**
 * void splitfloat(u32 val,char**up,char**down,int*sign,int*exp,int*man)
 *
 * converts a float into sign, exponent and mantissa
 *
 * val			the float to convert
 * sign,exp,man	pointerreferences to the destination integers
 *
 */
void splitfloat(u32 val,char**up,char**down,int*sign,int*exp,int*man)
{
	*sign= (val>>31);
	*exp= ((val>>23)&255)-127;
	*man=  ((val)&((1<<23)-1))+(1<<23);

	*up=splitbuf;

	char*sb=splitbuf;
	if(*sign) *sb++='-';
	if(*exp<0)
	{
		sb+=sprintf(sb,"%d",man)-1;
	}
	else
	{
		sb+=sprintf(sb,"ops")-1;
	}
	
	*down=splitbuf;
	splitbuf[0]=0;
}
/**
 * void splitfloat(u32 val,char**up,char**down,int*sign,int*exp,int*man)
 *
 * converts a double into sign, exponent and mantissa
 *
 * val			the double to convert
 * sign,exp,man	pointerreferences to the destination integers/quadwords
 *
 */
void splitdouble(u64 val,char**up,char**down,int*sign,int*exp,u64*man)
{
	u32 val1,val2;
	eight2four(val,&val1,&val2);

	*sign= (val2>>31);
	*exp= ((val2>>20)&2047)-1023;
	val2&=((1<<20)-1);
	four2eight(man,val1,val2);

	*up=splitbuf;
	*down=splitbuf;
	splitbuf[0]=0;
}

